#include "register.h"

static char str_true[] = "true";
static char str_false[] = "false";

static struct cr0_state parse_cr0(uint32_t val) {
	return (struct cr0_state){
		.protection_enable   = (val >>  0) & 1,
		.monitor_coprocessor = (val >>  1) & 1,
		.emulation           = (val >>  2) & 1,
		.task_switched       = (val >>  3) & 1,
		.extension_type      = (val >>  4) & 1,
		.numeric_error       = (val >>  5) & 1,
		.write_protect       = (val >> 16) & 1,
		.alignment_mask      = (val >> 18) & 1,
		.no_write_through    = (val >> 29) & 1,
		.cache_disable       = (val >> 30) & 1,
		.paging              = (val >> 31) & 1,
	};
}

struct cr0_state get_cr0(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr0\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return parse_cr0(val);
}

static char cr0_print_format[] =
	"CR0:\n"
	" - protection_enable: %s\n"
	" - monitor_coprocessor: %s\n"
	" - emulation: %s\n"
	" - task_switched: %s\n"
	" - extension_type: %s\n"
	" - numeric_error: %s\n"
	" - write_protect: %s\n"
	" - alignment_mask: %s\n"
	" - no_write_through: %s\n"
	" - cache_disable: %s\n"
	" - paging: %s\n";

int sprintf_cr0(struct cr0_state s, char *buf, ssize_t n) {
	return snprintf(
		buf, n, cr0_print_format,
		s.protection_enable   ? str_true : str_false,
		s.monitor_coprocessor ? str_true : str_false,
		s.emulation           ? str_true : str_false,
		s.task_switched       ? str_true : str_false,
		s.extension_type      ? str_true : str_false,
		s.numeric_error       ? str_true : str_false,
		s.write_protect       ? str_true : str_false,
		s.alignment_mask      ? str_true : str_false,
		s.no_write_through    ? str_true : str_false,
		s.cache_disable       ? str_true : str_false,
		s.paging              ? str_true : str_false
	);
};

uint64_t get_cr1(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr1\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return val;
}

uint64_t get_cr2(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr2\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return val;
}

static struct cr3_state parse_cr3(uint64_t val) {
	return (struct cr3_state){
		.page_level_write_through = (val >>  3) & 1,
		.page_level_cache_disable = (val >>  4) & 1,
		.page_directory_base      = (void *)(val & 0xFFFFFFFFFFFFF000),
	};
}

struct cr3_state get_cr3(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr3\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return parse_cr3(val);
}

void set_cr3(uint64_t val) {
	asm(
		".intel_syntax\n"
		"mov %%cr3, %0\n"
		".att_syntax\n"
		:
		: "r" (val)
		:
	);
}

static char cr3_print_format[] =
	"CR3:\n"
	" - page_level_write_through: %s\n"
	" - page_level_cache_disable: %s\n"
	" - page_directory_base: 0x%p\n";

int sprintf_cr3(struct cr3_state s, char *buf, ssize_t n) {
	return snprintf(
		buf, n, cr3_print_format,
		s.page_level_write_through ? str_true : str_false,
		s.page_level_cache_disable ? str_true : str_false,
		s.page_directory_base
	);
};

static struct cr4_state parse_cr4(uint32_t val) {
	return (struct cr4_state){
		.virtual_mode_extensions           = (val >>  0) & 1,
		.protected_mode_virtual_interrupts = (val >>  1) & 1,
		.time_stamp_disable                = (val >>  2) & 1,
		.debugging_extensions              = (val >>  3) & 1,
		.page_size_extensions              = (val >>  4) & 1,
		.physical_address_extension        = (val >>  5) & 1,
		.machine_check_enable              = (val >>  6) & 1,
		.page_global_enable                = (val >>  7) & 1,
		.perf_mon_counter_enable           = (val >>  8) & 1,
		.os_unmasked_simd_float_exceptions = (val >> 10) & 1,
		.user_mode_instruction_prevention  = (val >> 11) & 1,
		.linear_addres_57_bit              = (val >> 12) & 1,
		.vmx_enable                        = (val >> 13) & 1,
		.smx_enable                        = (val >> 14) & 1,
		.fsgbase_enable                    = (val >> 16) & 1,
		.pcid_enable                       = (val >> 17) & 1,
		.extended_states_enable            = (val >> 18) & 1,
		.key_locker_enable                 = (val >> 19) & 1,
		.smep_enable                       = (val >> 20) & 1,
		.smap_enable                       = (val >> 21) & 1,
		.user_mode_page_protection_keys    = (val >> 22) & 1,
		.control_flow_enforcement_enable   = (val >> 23) & 1,
		.supervisor_page_protection_keys   = (val >> 24) & 1,
	};
}

struct cr4_state get_cr4(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr4\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return parse_cr4(val);
}

static char cr4_print_format[] =
	"CR4:\n"
	" - virtual_mode_extensions: %s\n"
	" - protected_mode_virtual_interrupts: %s\n"
	" - time_stamp_disable: %s\n"
	" - debugging_extensions: %s\n"
	" - page_size_extensions: %s\n"
	" - physical_address_extension: %s\n"
	" - machine_check_enable: %s\n"
	" - page_global_enable: %s\n"
	" - perf_mon_counter_enable: %s\n"
	" - os_unmasked_simd_float_exceptions: %s\n"
	" - user_mode_instruction_prevention: %s\n"
	" - linear_addres_57_bit: %s\n"
	" - vmx_enable: %s\n"
	" - smx_enable: %s\n"
	" - fsgbase_enable: %s\n"
	" - pcid_enable: %s\n"
	" - extended_states_enable: %s\n"
	" - key_locker_enable: %s\n"
	" - smep_enable: %s\n"
	" - smap_enable: %s\n"
	" - user_mode_page_protection_keys: %s\n"
	" - control_flow_enforcement_enable: %s\n"
	" - supervisor_page_protection_keys: %s\n";

int sprintf_cr4(struct cr4_state s, char *buf, ssize_t n) {
	return snprintf(
		buf, n, cr4_print_format,
		s.virtual_mode_extensions                ? str_true : str_false,
		s.protected_mode_virtual_interrupts      ? str_true : str_false,
		s.time_stamp_disable                     ? str_true : str_false,
		s.debugging_extensions                   ? str_true : str_false,
		s.page_size_extensions                   ? str_true : str_false,
		s.physical_address_extension             ? str_true : str_false,
		s.machine_check_enable                   ? str_true : str_false,
		s.page_global_enable                     ? str_true : str_false,
		s.perf_mon_counter_enable                ? str_true : str_false,
		s.os_unmasked_simd_float_exceptions      ? str_true : str_false,
		s.user_mode_instruction_prevention       ? str_true : str_false,
		s.linear_addres_57_bit                   ? str_true : str_false,
		s.vmx_enable                             ? str_true : str_false,
		s.smx_enable                             ? str_true : str_false,
		s.fsgbase_enable                         ? str_true : str_false,
		s.pcid_enable                            ? str_true : str_false,
		s.extended_states_enable                 ? str_true : str_false,
		s.key_locker_enable                      ? str_true : str_false,
		s.smep_enable                            ? str_true : str_false,
		s.smap_enable                            ? str_true : str_false,
		s.user_mode_page_protection_keys         ? str_true : str_false,
		s.control_flow_enforcement_enable        ? str_true : str_false,
		s.supervisor_page_protection_keys        ? str_true : str_false
	);
};

uint64_t get_cr8(void) {
	uint64_t val = 0;

	asm(
		".intel_syntax\n"
		"mov %0, %%cr8\n"
		".att_syntax\n"
		: "=r" (val)
		:
		:
	);

	return val;
}

void set_wp(bool state) {
	pr_info("set_wp(%s)\n", state?"true":"false");

	if(state) {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr0\n"
			"or %%rax, 0b10000000000000000\n"
			"mov %%cr0, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	} else {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr0\n"
			"and %%rax, 0xFFFFFFFFFFFEFFFF\n"
			"mov %%cr0, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	}
	
	return;
}

void set_smap(bool state) {
	pr_info("set_smap(%s)\n", state?"true":"false");

	if(state) {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr4\n"
			"or %%rax, 0b1000000000000000000000\n"
			"mov %%cr4, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	} else {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr4\n"
			"and %%rax, 0xFFFFFFFFFFDFFFFF\n"
			"mov %%cr4, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	}
	
	return;
}

void set_smep(bool state) {
	pr_info("set_smep(%s)\n", state?"true":"false");

	if(state) {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr4\n"
			"or %%rax, 0b1000000000000000000000\n"
			"mov %%cr4, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	} else {
		asm(
			".intel_syntax\n"
			"mov %%rax, %%cr4\n"
			"and %%rax, 0xFFFFFFFFFFEFFFFF\n"
			"mov %%cr4, %%rax\n"
			".att_syntax\n"
			:
			:
			: "rax"
		);
	}
	
	return;
}
